/*
 * A simple MPI program to compute matrix multiplication.
 * 
 * Author: Scott Beca (sbeca), 267 362
 * Version: 1.1
 * Creation Date: 16/04/2011
 */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define ROOT 0
#define TAG 0

// The range of random numbers can be from -RANDOM_RANGE to +RANDOM_RANGE
#define RANDOM_RANGE 10

void masterProcess(int numProcs);
void workerProcess(int numProcs);
int** createMatrix(int rows, int cols);
void freeMatrix(int **matrix);

int main(int argc, char **argv)
{
    int numProcs;
    int myID;
    
    // All MPI programs start with MPI_Init
    MPI_Init(&argc, &argv);

    // Find out number of processes and this process's ID
    MPI_Comm_size(MPI_COMM_WORLD,&numProcs);
    MPI_Comm_rank(MPI_COMM_WORLD, &myID);
    
    // The ROOT rank becomes the master and all other ranks become workers
    if(myID == ROOT)
    {
        masterProcess(numProcs);
    }
    else
    {
        workerProcess(numProcs);
    }

    // End an MPI Program with MPI Finalize    
    MPI_Finalize();
    
    return 0;
}

void masterProcess(int numProcs)
{
    int matrixSize = 0;
    int i, x, y;
    MPI_Status stat;
    int *rows;
    int **mA, **mB, **mC;
    double time1, time2;
    FILE *outputFile;
    int leftovers;
    int part;
    
    rows = (int *)malloc(sizeof(int) * numProcs);
    
    printf("Enter the matrix size: ");
    fflush(stdout);
    scanf("%d", &matrixSize);
    printf("\n\n");
    fflush(stdout);
    
    while (matrixSize < 1)
    {
          printf("That matrix size is invalid.\n");
          printf("Enter the matrix size: ");
          scanf("%d", &matrixSize);
          printf("\n\n");
          fflush(stdout);
    }
    
    mA = createMatrix(matrixSize, matrixSize);
    mB = createMatrix(matrixSize, matrixSize);
    mC = createMatrix(matrixSize, matrixSize);
    
    // If the size of the matrix is 5 or less, allow the user to enter in the
    // values for the 2 matrices. If the size is greater than 5, assign random
    // values to the elements of matrix A and B.
    if (matrixSize <= 5)
    {
        printf("Enter rows and columns of First (A) matrix:\n");
        fflush(stdout);
        for(y = 0; y < matrixSize; y++)
            for(x = 0; x < matrixSize; x++)
                scanf("%d",&mA[y][x]);
        
        printf("Enter rows and columns of Second (B) matrix:\n");
        fflush(stdout);
        for(y = 0; y < matrixSize; y++)
            for(x = 0; x < matrixSize; x++)
                scanf("%d",&mB[y][x]);
    }
    else
    {
        srand(time(NULL));
        
        for(y = 0; y < matrixSize; y++)
            for(x = 0; x < matrixSize; x++)
                mA[y][x] = rand() % (RANDOM_RANGE * 2 + 1) - RANDOM_RANGE;
        
        for(y = 0; y < matrixSize; y++)
            for(x = 0; x < matrixSize; x++)
                mB[y][x] = rand() % (RANDOM_RANGE * 2 + 1) - RANDOM_RANGE;
    }
    
    // Save matrix A out to a file
    outputFile = fopen("./A.txt", "w");
    for(y = 0; y < matrixSize; y++)
    {
        for(x = 0; x < matrixSize; x++)
            fprintf(outputFile, "%d\t", mA[y][x]);
        fprintf(outputFile, "\n");
    }
    fclose(outputFile);

    // Save matrix B out to a file
    outputFile = fopen("./B.txt", "w");
    for(y = 0; y < matrixSize; y++)
    {
        for(x = 0; x < matrixSize; x++)
            fprintf(outputFile, "%d\t", mB[y][x]);
        fprintf(outputFile, "\n");
    }
    fclose(outputFile);

    // Work out how many rows each worker will process.
    rows[0] = 0;
    leftovers = matrixSize % (numProcs - 1);
    for (i = 1; i < numProcs; i++)
    {
        rows[i] = matrixSize / (numProcs - 1);
        
        if (leftovers > 0)
        {
            rows[i] += 1;
            leftovers--;
        }
    }
    
    // Start time
    time1 = MPI_Wtime();
    
    // Send matrix size to all workers
    MPI_Bcast((void *)&matrixSize, 1, MPI_INT, ROOT, MPI_COMM_WORLD);

    // Send matrix B to all workers
    MPI_Bcast((void *)mB[0], matrixSize * matrixSize, MPI_INT, ROOT, MPI_COMM_WORLD);
    
    // Send each worker a part of matrix A for the calculation
    part = 0;
    for (i = 1; i < numProcs; i++)
    {
        MPI_Send((void *)&rows[i], 1, MPI_INT, i, TAG, MPI_COMM_WORLD);
        MPI_Send((void *)mA[part], matrixSize * rows[i], MPI_INT, i, TAG,
                    MPI_COMM_WORLD);
    
        part += rows[i];
    }
    
    // Receive each part of matrix C from the workers
    part = 0;
    for (i = 1; i < numProcs; i++)
    {
        MPI_Recv((void *)mC[part], matrixSize * rows[i], MPI_INT, i, TAG,
                    MPI_COMM_WORLD, &stat);
    
        part += rows[i];
    }
    
    // End time
    time2 = MPI_Wtime();
    
    // Save matrix C out to a file
    outputFile = fopen("./C.txt", "w");
    for(y = 0; y < matrixSize; y++)
    {
        for(x = 0; x < matrixSize; x++)
            fprintf(outputFile, "%d\t", mC[y][x]);
        fprintf(outputFile, "\n");
    }
    fclose(outputFile);
    
    // Done! Output results in the form:
    // Matrix Size          Number of process             Time taken
    printf("%d\t%d\t%f\n", matrixSize, numProcs, time2 - time1);
    fflush(stdout);

    freeMatrix(mA);
    freeMatrix(mB);
    freeMatrix(mC);

    return;
}

void workerProcess(int numProcs)
{
    int x, y, i;
    int matrixSize;
    int rows;
    int **mA, **mB, **mC;
    MPI_Status stat;

    // Receive matrix size
    MPI_Bcast((void *)&matrixSize, 1, MPI_INT, ROOT, MPI_COMM_WORLD);

    mB = createMatrix(matrixSize, matrixSize);
    
    // Receive matrix B
    MPI_Bcast((void *)mB[0], matrixSize * matrixSize, MPI_INT, ROOT, MPI_COMM_WORLD);
    
    // Receive how many rows this worker is working on
    MPI_Recv((void *)&rows, 1, MPI_INT, ROOT, TAG, MPI_COMM_WORLD, &stat);

    mA = createMatrix(rows, matrixSize);
    mC = createMatrix(rows, matrixSize);
    
    // Receive this workers part of matrix A
    MPI_Recv((void *)mA[0], matrixSize * rows, MPI_INT, ROOT, TAG, MPI_COMM_WORLD,
                &stat);

    // Compute the result matrix
    for(y = 0; y < rows; y++)
    {
        for(x = 0; x < matrixSize; x++)
        {
            mC[y][x] = 0;
            for(i = 0; i < matrixSize; i++)
                mC[y][x] += mA[y][i] * mB[i][x];
        }
    }
    
    // Send computed matrix back to master
    MPI_Send((void *)mC[0], matrixSize * rows, MPI_INT, ROOT, TAG, MPI_COMM_WORLD);
    
    return;
}

int** createMatrix(int rows, int cols)
{
    int *colsArray;
    int **rowsArray;
    int i;
    
    colsArray = (int *)malloc(sizeof(int) * rows * cols);
    rowsArray = (int **)malloc(sizeof(int *) * rows);
    
    for (i = 0; i < rows; i++)
        rowsArray[i] = colsArray + (i * cols);

    return rowsArray;
}

void freeMatrix(int **matrix)
{
    free(matrix);
}
